// These classes are used to descibe how objects interact with one another.

#ifndef INTERACTION_H
#define INTERACTION_H

#include "object.h"

#define UNIVERSAL_GRAVITATIONAL_CONSTANT ( 6.67258e-11 )
#define DEFAULT_K ( 1.00 )
#define DEFAULT_EQ_POINT ( 0.00 )
#define DEFAULT_ELASTIC_LIMIT ( 0.00 )

// An abstract base class for dealing with two objects interacting
class interaction
{
	private:
	
	protected:
	object* _obj_1;
	object* _obj_2;
	
	force_id _force_1;
	force_id _force_2;
	
	
	public:

	interaction(void);
	interaction(object* One, object* Two);
	virtual ~interaction(void);
	
	virtual void calculate(void) {};
	virtual bool link(object* One, object* Two);
};

class gravitation : public interaction
{
	private:
	// Universal gravitational constant
	static scalar _G;
	
	public:
	gravitation(object* One, object* Two);
	
	void calculate(void);
	scalar G(void) const;
	bool setG(scalar G);
};

class hookes_law : public interaction
{
	private:
	// spring constant
	scalar _k;
	// equilibrium distance... stretching or compressing starts here
	scalar _eq_point;
	// Limit of elasticity.  When a spring stretches farther than this,
	// it breaks.  No limit if _limit == 0.0
	scalar _limit;
	// Need to keep track of what universe we're in, which
	// node we're in.
	class universe* _parent_uni;
	spring_id _id;
	
	public:
	hookes_law(object* One, object* Two, scalar k = DEFAULT_K,
	           scalar x = DEFAULT_EQ_POINT,
	           scalar max_x = DEFAULT_ELASTIC_LIMIT);
	hookes_law(class universe* Parent,  spring_id Id,
	           scalar k = DEFAULT_K,
	           scalar x = DEFAULT_EQ_POINT, 	
                   scalar max_x = DEFAULT_ELASTIC_LIMIT);
	hookes_law(class universe* Parent, spring_id Id,
	           object* One, object* Two, scalar k = DEFAULT_K,
	           scalar x = DEFAULT_EQ_POINT,
	           scalar max_x = DEFAULT_ELASTIC_LIMIT);
	~hookes_law(void);
	
	void calculate(void);
	scalar k(void) const;
	bool set_k(scalar k);
	scalar eq(void) const;
	bool setEq(scalar x);
	scalar limit(void) const;
	bool setLimit(scalar max_x);
	bool setId(spring_id Id);
	
};

// Damping, i.e. friction and fluid resistance.  User supplies coefficients
// for equation F = -(a*v^2 + b*v + c).  C is frictional force, a & b are
// quadratic and linear proportionality constants.  Resistance always
// related to velocity.
class damping : public interaction
{
	private:

	// Damping coefficients
	scalar _a, _b, _c;

	public:
	damping(scalar c, scalar b = 0.0, scalar a = 0.0);
	damping(object* One, scalar c, scalar b = 0.0, scalar a = 0.0);

	void calculate(void);
	vector coeff(void);
	bool setCoeff(vector K);

};

// Collisions.  Should trigger a "collide" event whenever the objects'
// separation is < their combined radii.  If either radius is zero, ignore.
class collision : public interaction
{

	private:
	class universe* _parent_uni;
//	solve by making void pointers?  not good, i say.
//	point_mass* Mass_1;
//	point_mass* Mass_2;
	scalar Last_Collision;

	public:

	// Collisions can't exist without universe: must use point_mass
	// type.  Therefore, no default constructor.
	collision(universe* This);
	collision(universe* This, object* One, object* Two);

	void calculate(void);
};

#endif