// Physical objects in the RockSpace universe.
// An object will contain properties for a single object that is not being
// influenced by any other objects.  Save interactions for the "universe"
// class.

#ifndef OBJECT_H
#define OBJECT_H

#include "vector.h"

// Force is a vector quantity; resultant force is the vector sum of every
// force acting upon an object.  Store as a linked list.
typedef short force_id;

struct FORCE
{
	vector data;
	struct FORCE* next;
	force_id id;
};

typedef struct FORCE force;

class object
{
	private:
	
	char *_name;
	scalar _mass, _init_time;
	vector _velocity, _position;
	scalar _time_multiplier;
	bool _paused;
	
	// Store list of concurrent forces, as well as their vector sum.
	// Only recalculate sum if the forces have changed.
	force *_force_list;
	vector _resultant;

	// Need a system-independant way to determine time w/ millisecond
	// accuracy.
	scalar _now(void) const;
		
	// dealing with the list of forces
	bool _remove_all_forces(void);
	
	// Reset timer to zero, initial s & v.
	bool _reset_conditions(void);
	
	
	// Perhaps implement rotational (angular) stuff later?
	
	public:

	// Overloaded set of constructors
        object(void);
        object(char *Name,
           scalar Mass = 0.0,
           vector Position = ToVector(0.0, 0.0, 0.0),
           vector Velocity = ToVector(0.0, 0.0, 0.0));
	virtual ~object(void);
		
	// Interface methods for getting info about the object
	char* name(void) const;
	scalar mass(void) const;
	vector position(void) const;
	vector velocity(void) const;
	vector acceleration(void) const;
	scalar time(void) const;
	
	// Interface methods for altering the object
	bool setName(char* New_Name);
	bool setMass(scalar New_Mass);
	bool setPosition(vector New_Pos);
	bool setVelocity(vector New_Vel);
	bool setAcceleration(vector New_Accel);
	bool setTime(scalar New_T);
	
	// The force interface
	force_id applyForce(vector F);
	bool removeForce(force_id Old_F);
	vector getForce(force_id Id);
	bool setForce(force_id Id, vector F);
	force_id* listForces(void);
		
	// Fun with time.  Good for special effects, maintaining universe.
	scalar timeMult(void);
	bool setTimeMult(scalar New_X);
	bool pause(void);
	bool setPause(bool Pause);
	bool setPause(void);
	bool setUnPause(void);
	bool togglePause(void);
	
	// Hooks for later application
	virtual force_id id(void) {return 0;};
};

#endif