// Physics laws and interactions.

#ifndef PHYSICS_H
#define PHYSICS_H

typedef short point_mass_id;
typedef point_mass_id object_id;

typedef short interaction_id;
typedef interaction_id spring_id;

#include "object.h"
#include "interaction.h"

#define DEFAULT_TIME_X (1.00)


// container class for making linked lists of interactions.
struct INTERACT_LIST
{
	// Must be pointer so we can select type later.
	interaction* data;
	struct INTERACT_LIST* next;
	object_id members[2];
	interaction_id id;
};
typedef struct INTERACT_LIST interaction_node;



// Here's a structure to help the universe manage objects with mass and
// such.  We'll take an object and add hooks to make it easier to play
// with others.
class point_mass : public object
{
	private:
	point_mass_id _id;
	interaction_node* _gravity;
	damping* _friction;
	point_mass* _next;
	point_mass* _prev;
	// For collision detection
	scalar _radius;
	interaction_node* _object_borders;
	// A user-definable data repository
	void* _errata;
	
	public:
        point_mass(void);
        point_mass(char *Name,
           scalar Mass = 0.0,
           vector Position = ZERO_VECTOR,
           vector Velocity = ZERO_VECTOR,
           scalar r = 0.0);
	~point_mass(void);
	
	point_mass* next() const;
	bool setNext(point_mass* Node);
	
	point_mass_id id(void) const;
	bool setId(point_mass_id Id);
	
	scalar radius(void) const;

	void* userData(void) const;
	bool setUserData(void* Data_Ptr);
		
	friend class universe;
        friend void Gravity_Link(point_mass* New_Obj,
                                  point_mass* Existant);
        friend void Refresh_Gravity(point_mass* Object);
        friend void Update_Friction(point_mass* Object);
        friend void Refresh_Collision(point_mass* Object);
        friend void Collision_Link(universe* Uni, point_mass* New_Obj,
                    point_mass* Existant);

};

// Here are some control structures, useful for returning information from
// the universe
struct VECTOR_ID_BLOCK
{
	vector data;
	object_id id;
};

typedef struct VECTOR_ID_BLOCK multi_vector;

struct SCALAR_ID_BLOCK
{
	scalar data;
	object_id id;
};

typedef struct SCALAR_ID_BLOCK multi_scalar;

// A structure used to keep track of all the spring links in the universe
struct SPRING_ID_BLOCK
{
	object_id one, two;
	hookes_law* data;
};

typedef struct SPRING_ID_BLOCK multi_spring;

// The main idea here is to give a direct & easy way to apply a controlling
// force to an object, without going through physical laws.
class object_control
{
	private:
	object* _obj;
	object_id _id;
	force_id _force;
	
	public:
	object_control(void);
	object_control(object* Obj);
	~object_control(void);
	
	bool control(object* Obj);
	bool setForce(vector F);
	vector forceIs(void) const;
};

// Here's the universe.  This will administer all the objects created.
class universe
{
	private:
	point_mass* _object_list;
	interaction_node* _spring_list;
	scalar _time_x;
	scalar _initial_t;
	bool _paused;
	// damping coefficients, arranged (a, b, c).
	// same order used in eq., opposite from use in link constructor.
	scalar _damping_coeff[3];
	
	// useful tools
	point_mass* _get_by_id(point_mass_id Id) const;
	scalar _now(void) const;
	
	// Performs the supplied function on every object in the list.
	void _on_every_object(point_mass* List, void (*fn)(point_mass*));
	void _on_every_object(interaction_node* List,
	                      void (*fn)(interaction_node*));
	// Generic algorithm.  Returns address of object where fn(Obj, Arg)
	// is true.  Returns NULL if nothing works.
	point_mass* _search(point_mass* List, void* Arg,
	                    bool (*fn)(point_mass*, void*)) const;
	
	public:
	universe(void);
	~universe(void);
	
		
	// Populating the universe
	object_id add(scalar Mass = 0.0,
	   vector Position = ZERO_VECTOR,
           vector Velocity = ZERO_VECTOR,
           scalar Radius = 0.0);

        bool remove(object_id Target);
        bool removeAll(void);

        // Tag maintenance
        char* tag(object_id Target);
        bool setTag(object_id Target, char* Tag);

        // Arbitrary info from each object
	void* userData(object_id Target);
	bool setUserData(object_id Target, void* Data_Ptr);

        // Gravitation
	scalar G(void) const;
	bool setG(scalar New_G);
	bool calculateGravity(void);
	
	// Springs.  Good for connecting objects.
	spring_id springLink(object_id One, object_id Two,
	                     scalar k = DEFAULT_K,
	                     scalar x = DEFAULT_EQ_POINT,
	                     scalar max_x = DEFAULT_ELASTIC_LIMIT);
	bool springUnLink(spring_id Target);
	// Calculates all springs
	bool calculateSpring(void);
	bool calculateSpring(spring_id Target);
	// Returns all springs and the objects they work on
	multi_spring* springs(void);
	
	// Damping (friction, air resistance)
	// Neccesary for realistic motion.
	vector dampingCoeff(void);
	vector dampingCoeff(object_id Id);
        bool setDampingCoeff(vector K);
        bool setDampingCoeff(object_id Id, vector K);
        bool calculateDamping(void);
        bool calculateDamping(object_id Id);
	
        // Collisions.  Based on conservation of momentum.
        bool collide(object_id One, object_id Two,
                     scalar Elasticity = 1.00);
	// Automatic collision detection
	bool calculateCollision(void);
	bool calculateCollision(object_id Id);
        			
	// Querying objects.
	bool exists(object_id Id) const;
	vector position(object_id Id) const;
	multi_vector* position(void) const;
	vector velocity(object_id Id) const;
	multi_vector* velocity(void) const;
	vector acceleration(object_id Id) const;
	multi_vector* acceleration(void) const;
	scalar mass(object_id Id) const;
	multi_scalar* mass(void) const;
	scalar radius(object_id Id) const;
	
	// Manipulating objects
	bool setPosition(object_id Id, vector S);
	bool setVelocity(object_id Id, vector V);
	bool setAcceleration(object_id Id, vector A);
	bool setMass(object_id Id, scalar M);
	
	// I think this will be quite useful.
	object_control control(object_id Id);
	
	// Fun with time.
	bool pause(void) const;
	bool pause(object_id Id) const;
	bool setPause(bool Pause);
	bool setPause(object_id Id, bool Pause);
	bool togglePause(void);
	bool togglePause(object_id Id);

	// Returns time in seconds from creation of the universe.
	scalar time(void) const;
		
	scalar timeMult(void) const;
	scalar timeMult(object_id Id) const;
	bool setTimeMult(scalar New_X);
	bool setTimeMult(object_id Id, scalar New_X);
	
	
	
};


#endif