/* forces.cc - create and maintain a list of forces.
        Force is a vector quantity.  The resultant force an object feels is
   equal to the vector sum of all individual forces acting upon it.  We shall
   store all forces in a linked list, calculating the resultant force when
   needed.  This will hopefully improve performance.
*/

#include "object.h"

// for NULL
#include <stddef.h>

// helper functions used locally
force* Find_Tail(force* Here);
bool Delete_Id(force* Here, force_id Target);
vector Sum_Forces(force* Here);
void Delete_Forces(force* Here);
int Count_Forces(force* Here);
force* Find_By_Id(force* Here, force_id Id);


// Add a new force to the list, return unique id #
force_id object::applyForce(vector F)
{
	force* Tail = Find_Tail(_force_list);
        int Last_Id = 0;

        _reset_conditions();

	if (Tail == NULL) {
		// List is empty
		_force_list = Tail = new force;
	} else {
		Tail->next = new force;
		Last_Id = Tail->id;
		Tail = Tail->next;
	}
	
	Tail->data = F;
	Tail->next = NULL;
	// Will this always yield a unique #?  i think so, barring wraparound
	Tail->id = Last_Id + 1;
	
	_resultant = Sum_Forces(_force_list);
	
	return Tail->id;
}

force* Find_Tail(force* Here)
{
	if ((Here == NULL) || (Here->next == NULL)) {
		return Here;
	} else {
		return Find_Tail(Here->next);
	}
}

bool object::removeForce(force_id Old_F)
{
	bool List_Tainted = false;
	
	// This call is not needed if no force is deleted.  How to implement?
	_reset_conditions();
	
	// Let's test for special cases _before_ we send it to the
	// recursive delete thing.  This might just be simpler!
	if (_force_list == NULL) {
		// no forces, nothing to do.
		return false;
	} else if (_force_list->next == NULL) {
		// only one force in list.
		if (_force_list->id == Old_F) {
			delete _force_list;
			_force_list = NULL;
			
			// Signal we've modified force list
			List_Tainted = true;
			
		}
	} else {
		List_Tainted = Delete_Id(_force_list, Old_F);
	}
		
	if (List_Tainted) {
		_resultant = Sum_Forces(_force_list);
	}
	
	return List_Tainted;
	
}

// Search through the list recursively, trim when found.
bool Delete_Id(force* Here, force_id Target)
{
	if ((Here == NULL) || (Here->next == NULL)) {
		// the end of the line
		return false;
	} else if (Here->next->id == Target) {
		force* Temp = Here->next->next;
		delete Here->next;
		Here->next = Temp;
		return true;
	} else {
		return Delete_Id(Here->next, Target);
	}
}


// wee-hee.  recursion is fun.
vector Sum_Forces(force* Here)
{
	if (Here == NULL) {
		return ZERO_VECTOR;
	} else {
		return (Here->data + Sum_Forces(Here->next));
	}
}

bool object::_remove_all_forces(void)
{
	if (_force_list == NULL) {
		return false;
	} else {
		Delete_Forces(_force_list);
		_force_list = NULL;
		return true;
	}
}

void Delete_Forces(force* Here)
{
	if (Here == NULL)
		return;
	
	Delete_Forces(Here->next);
	delete Here;
}

// Here, modify a force given by its ID.
bool object::setForce(force_id Id, vector F)
{
	force* Target = Find_By_Id(_force_list, Id);
	if (Target == NULL) {
		return false;
	}
	
	Target->data = F;
	// Update force for kinematics.
	_reset_conditions();
	_resultant = Sum_Forces(_force_list);
	return true;
	
}

force* Find_By_Id(force* Here, force_id Id)
{
	if (Here == NULL) {
		return NULL;
	} else if (Here->id == Id) {
		return Here;
	} else {
		return Find_By_Id(Here->next, Id);
	}
}

// Here are some useful functions for gathering information about the forces

// simply get the value of a force
vector object::getForce(force_id Id)
{
	force* F = Find_By_Id(_force_list, Id);
	if (F == NULL) {
		return ZERO_VECTOR;
	}
	
	return F->data;
}

// Return a null-terminated array of all force id's.  leave to caller to
// delete when finished.
force_id* object::listForces(void)
{
	int Number = Count_Forces(_force_list);
	force_id* Result = new force_id[Number + 1];
	force* Here = _force_list;
	int i = 0;

	while (Here != NULL) {
		Result[i++] = Here->id;
		Here = Here->next;
	}
	
	// Terminate array
	Result[i] = (force_id) NULL;
	return Result;
}

int Count_Forces(force* Here)
{
	static int Count = 0;
	if (Here == NULL) {
		return Count;
	} else {
		Count++;
		return Count_Forces(Here->next);
	}
}