/* kinematics.cc - vector kinematics calculations
        "Kinematics" is the description of an objects motion.  This means
   that knowing an object's initial conditions, we can predict its future
   state.  As a result, these equations are only valid for when the initial
   conditions remain valid.
        My equations assume constant force over the time interval.  This is
   great for gravitational-type stuff, less good for control dynamics, etc.
   They'll work, but they are inefficient because the resultant force must
   be recalculated every time a force is added or removed.  However, more
   efficient methods for calculating force require some knowledge of the
   type of interaction.  This can't be done without knowledge of the rest of
   the universe.
*/

#include "object.h"

scalar object::mass(void) const
{
	return _mass;
}

// Should I add error checking for objects with no mass?

// Calculate position, using initial conditions
// S = So + v*t + 1/2*a*t^2
vector object::position(void) const
{
	
	scalar T = time();
	
	return (_position
	        + (T * _velocity)
	        + ((T * T / _mass / 2.0) * _resultant));
	
}

vector object::velocity(void) const
{
	return (time()/_mass) * _resultant + _velocity;
}

// Remember newton's 2nd law: F = ma.
vector object::acceleration(void) const
{
	return (_resultant / _mass);
}

// Methods for setting the parameters of the object.  Each one of these will
// cause a reset of the timer clock, execpt for mass.  (expensive?)

// I'm intentionaly being lenient when assigning values.  Negative mass will
// be very interesting.
bool object::setMass(scalar New_Mass)
{
	_mass = New_Mass;
	return true;
}

bool object::setPosition(vector New_Pos)
{
	// _reset_conditions also resets position, so in effect,
	// we change the value twice.  Change if this is too expensive.
 	_reset_conditions();
	_position = New_Pos;
	return true;
}

bool object::setVelocity(vector New_Vel)
{
	_reset_conditions();
	_velocity = New_Vel;
	return true;
}

// Remember, F = ma.  Adjust the resultant force to create desired
// acceleration.  You should probably avoid doing this when dealing with
// dynamic systems:  it clobbers all previously-applied forces.
bool object::setAcceleration(vector New_Accel)
{
	if (_mass == 0.0) {
		return false;  // no mass, no acceleration
	}
	
	_remove_all_forces();
	applyForce(_mass * New_Accel);
	return true;
}