/* time.cc - timing considerations.
        RockSpace strives to be a real-time physics engine.  To that end,
   there must be a method to determine time with millisecond precision.
   ANSI C can only determine real time down to the second... next to
   worthless for physics!
        I'll use the gettimeofday() function that comes with UNIX to
   determine time.  I'll port it to windows when I get around to it.
*/

#include "object.h"

#ifndef WIN32
// For gettimeofday()
#include <sys/time.h>
#include <unistd.h>
#else
#include <sys/timeb.h>
#endif

// Should return current time as a floating-point number of seconds.
scalar object::_now(void) const
{
#ifndef WIN32
	struct timeval Now;
	gettimeofday(&Now, NULL);
	
	return (Now.tv_sec + (1e-6 * Now.tv_usec));

#else
	struct timeb Now;
	ftime(&Now);
	return (Now.time + (1e-3 * Now.millitm));
#endif    //  not WIN32		

}

// Return the number of seconds elapsed since the initial conditions were
// true.
scalar object::time(void) const
{
	if (_paused) {
		return 0.0;
	}
	
	return (_time_multiplier * (_now() - _init_time));
}

// Tell the object to behave as if New_T seconds had elapsed since
// initializing.  Best used with a comparison to time().
bool object::setTime(scalar New_T)
{
	if (_paused) {
		return false;
	}
	
	if (_time_multiplier != 1.00) {
		_init_time = ((_time_multiplier * _now()) - New_T)
		             / _time_multiplier;
	} else {
		_init_time = _now() - New_T;
	}	
	return true;
}

// The time multiplier adjusts the rate of time atomically for each object!
// It'd be really easy to implement "Bullet Time" (tm) as in "The Matrix"
scalar object::timeMult(void)
{
	return _time_multiplier;
}

// When we reset, the time multiplier, we have to make the object believe
// the same amount of time has passed.  Therefore, we need to reset the
// initial time.
bool object::setTimeMult(scalar New_X)
{
	// For right now, the easiest way to reset the timer is to simply
	// set it to zero.
	_reset_conditions();
	
	// I'll allow negative multipliers.  Might be really cool...
	// If it doesn't work, it gets yanked.
	_time_multiplier = New_X;
	return true;
}
	
bool object::pause(void)
{
	return _paused;
}

bool object::setPause(bool Pause)
{
	if (Pause) {
		return setPause();
	} else {
		return setUnPause();
	}
}

bool object::setPause(void)
{
	// Store the present state attributes.
	_position = position();
	_velocity = velocity();

	// Now flag to pause the object	
	return (_paused = true);
}

bool object::setUnPause(void)
{
	// When we release the object, we want its initial time to be reset.
	_init_time = _now();
	
	return !(_paused = false);
}

// Flip pause state... return new value.
bool object::togglePause(void)
{
	if (_paused) {
		return !(setUnPause());
	}
	
	return setPause();
}

bool object::_reset_conditions(void)
{
	_position = position();
	_velocity = velocity();
        if (_paused) {
        	return false;
        }
		
	_init_time = _now();
	return true;
}
