/* collision_solver.cc -- elastic collisions
        I'll use this file to easily "plug in" new implementations of the
   elastic collision math.  I don't have anything that works really well,
   yet. I'll keep you posted.
*/

#include "physics.h"
#include <stddef.h>
#include <math.h>
#include <assert.h>

#define PI ( 3.14159265358979 )

#define ZERO_THRESHOLD ( 1e-4 )

static bool Solve_Quadratic(scalar a, scalar b, scalar c, scalar x[2]);

// debug
#include <stdio.h>

bool universe::collide(object_id One, object_id Two, scalar k)
{
	fprintf(stderr, "**---- universe::collide()\n");

	point_mass* Obj_1 = _get_by_id(One);
	point_mass* Obj_2 = _get_by_id(Two);

	if ((Obj_1 == NULL) || (Obj_2 == NULL)) {
		return false;
	}

	scalar t0 = time();
	/* cache positions for use later... trust me */
	vector s1 = Obj_1->position(), s2 = Obj_2->position();

	scalar* Result = Obj_1->velocity().array();

	fprintf(stderr, "Input: m1 = %g, m2 = %g\n", Obj_1->mass(),
	        Obj_2->mass());

	fprintf(stderr, "Input: v1 = (%g, %g, %g)\n", Result[0],
	        Result[1], Result[2]);
	Obj_2->velocity().array();
 	fprintf(stderr, "Input: v2 = (%g, %g, %g)\n", Result[0],
		Result[1], Result[2]);

	// Cache velocities for performance and accuracy
	vector v1 = Obj_1->velocity(), v2 = Obj_2->velocity();
	scalar m1 = Obj_1->mass(), m2 = Obj_2->mass();
	// array has 2 spaces for both results from the quadratic formula
	scalar x[2];

	// here's where we store new values for the velocity vectors
	scalar new_v1[3], new_v2[3];

	// perform calculations for each of the three dimensions independently
	for (int i = 0; i < 3; i++) {
		scalar vo = v1.access(i) - v2.access(i);
		fprintf(stderr, "Dim %d: vo = %g\n", i, vo);

		// special cases

		if (fabs(vo) < ZERO_THRESHOLD) {
			// They were both moving in the same direction
			fprintf(stderr, "Skipping dim %d: no collision.\n", i);
			new_v1[i] = v1.access(i);
			new_v2[i] = v2.access(i);
			continue;
		} else if (fabs(m1 / m2) < ZERO_THRESHOLD) {
			// m2 much larger than m1
			fprintf(stderr, "m2 much larger than m1\n");
			if (vo < 0.0) {
				// objs coming closer together
				new_v1[i] = -v1.access(i);
			} else {
				new_v1[i] = v1.access(i);
			}
			new_v2[i] = v2.access(i);
			continue;
		} else if (fabs(m2 / m1) < ZERO_THRESHOLD) {
			fprintf(stderr, "m1 much larger than m2\n");
			if (vo < 0.0) {
				// the objects are coming closer together
				new_v2[i] = -v2.access(i);
			} else {
				new_v2[i] = v2.access(i);
			}
			new_v1[i] = v1.access(i);
			continue;
		}


		// OK, now solve for the most general case.
		// We will simplify the equations by shifting the frame of
		// reference.  Basically, all velocities will be relative
		// to v2, meaning that all terms of v2 in the equations will
		// be zero and will drop out.  We must remember to shift back
		// to the original frame of reference at the end!

		if (Solve_Quadratic((m1 + (m1*m1/m2)),
		                    (-(2*m1*m1/m2*vo)),
				    ((m1*m1/m2) - m1),
				    x)) {
        		// Found real-valued results
			fprintf(stderr, "Found real results: (%g, %g)\n",
				x[0], x[1]);
			new_v1[i] = x[0];
			new_v2[i] = m1 * (vo - x[0]) / m2;
			// need to check if the results are valid
			if (  fabs(((m1 * new_v1[i] * new_v1[i])
			             + (m2 * new_v2[i] * new_v2[i]))
			           - (m1 * vo * vo))
			       > ZERO_THRESHOLD) {
          			// whoops. didn't match.  swap.
				new_v1[i] = x[1];
				new_v2[i] = m1 * (vo - x[1]) / m2;
    			}

			// finally, we need to correct for the frame of reference.
			new_v1[i] += v2.access(i);
			if (fabs(new_v1[i]) < ZERO_THRESHOLD) {
				new_v1[i] = 0.0;
			}
			new_v2[i] += v2.access(i);
			if (fabs(new_v2[i]) < ZERO_THRESHOLD) {
				new_v2[i] = 0.0;
			}
   		} else {
			// non-real-valued result
			fprintf(stderr, "Quadratic equation unsolvable!\n");
			abort();
		}
	}

	// finally, assign the new velocities to the objects
	Obj_1->setVelocity(vector(new_v1[0], new_v1[1], new_v1[2]));
	Obj_2->setVelocity(vector(new_v2[0], new_v2[1], new_v2[2]));

	Obj_1->velocity().array();
	fprintf(stderr, "Set: v1' = (%g, %g, %g)\n", Result[0],
	        Result[1], Result[2]);
	Obj_2->velocity().array();
 	fprintf(stderr, "Set: v2' = (%g, %g, %g)\n", Result[0],
		Result[1], Result[2]);

	// OK, we also want to adjust their positions, make sure that
	// an object will never be "inside" another.
	if ((s1-s2).mag() < (Obj_1->radius() + Obj_2->radius())) {
		// inside the other... move the one w/ smaller r
		fprintf(stderr, "relocating...\n");
		if (Obj_1->radius() < Obj_2->radius()) {
			Obj_1->setPosition(s2 + (((s1 - s2).dir())*
			       (Obj_1->radius()+Obj_2->radius())));
          	} else {
			Obj_2->setPosition(s1 + (((s2 - s1).dir())*
			       (Obj_1->radius()+Obj_2->radius())));
          	}
	}



	fprintf(stderr, "That took %g seconds.\n", time() - t0);
//	exit(1);
	return true;
}

// OK, specify two objects to collide.  k is the "coefficient of
// elasticity," as I call it.  It should be the percentage of kinetic energy
// which will be conserved after the collision, ranging from 0.00 to 1.00

/* For simplicity, the collision is modeled as happening instantaneously
   (in zero time).  As a result, we simply change the objects' velocities
   and leave their accelerations alone.  To calculate the new velocities,
   find the cartesian components of the velocity vectors and then solve the
   system of equations given by conservation of momentum and kinetic
   energy. Got it?  Good.
*/

/*bool old_universe__collide(object_id One, object_id Two, scalar k)
{
	fprintf(stderr, "======  universe::collide()\n");

	point_mass* Obj_1 = _get_by_id(One);
	point_mass* Obj_2 = _get_by_id(Two);

	if ((Obj_1 == NULL) || (Obj_2 == NULL)) {
		return false;
	}

	scalar* Result = Obj_1->velocity().array();

	fprintf(stderr, "Input: m1 = %g, m2 = %g\n", Obj_1->mass(),
	        Obj_2->mass());

	fprintf(stderr, "Input: v1 = (%g, %g, %g)\n", Result[0],
	        Result[1], Result[2]);
	Obj_2->velocity().array();
 	fprintf(stderr, "Input: v2 = (%g, %g, %g)\n", Result[0],
		Result[1], Result[2]);

	// Cache velocities for performance and accuracty
	vector v_1 = Obj_1->velocity(), v_2 = Obj_2->velocity();
	// array has 2 spaces for both results from the quadratic formula
	scalar x[2];

	// here's where we store new values for the velocity vectors
	scalar new_v1[3], new_v2[3];

	scalar M = Obj_1->mass() / Obj_2->mass();

	// repeat once for each dimension
	for (int i = 0; i < 3; i++) {
		// cache a bunch of intermediate calculations
		scalar v1 = v_1.access(i), v2 = v_2.access(i);
		scalar B = (M * v1) + v2;
		scalar KE = k * (Obj_1->mass() * v1*v1 + Obj_2->mass() * v2*v2);

		// Special cases
		// if both velocities are identical, then there is really no
		// collision.  detect this and then skip to the next dimension.
		if (fabs(v1 - v2) < ZERO_THRESHOLD) {
			fprintf(stderr, "Skipping dim %d: no collision.\n", i);
			new_v1[i] = v1;
			new_v2[i] = v2;
			continue;
		}

		if (Solve_Quadratic((M*M*Obj_2->mass() + Obj_1->mass()),
		                     2*M*B*Obj_2->mass(),
		                     Obj_2->mass()*B*B - KE,
		                     x)) {

			fprintf(stderr, "Results: For dimension %d v' = (%g, %g)\n",
				i, x[0], x[1]);

			// The answer is real-valued.  check if one is negative
			// Doesn't matter if it's negative!  velocity is a
			// vector quantity.
			//if (!(    (x[0] == 0 && x[1] == 0)
			//        || (x[0] >= 0 && x[1] < 0)
			//       || (x[0] < 0 && x[2] >= 0))) {
			//	fprintf(stderr, "Dim %d: Ambiguous answer!!\n", i);
			//}

			// finally, solve the simple momentum conservation
			// equation to get the new v2
			// First solve using the first value returned for v1.
			// Test if the original kinetic energy equation holds.
			// If not, use the second value.
			//
			new_v1[i] = x[0];
			new_v2[i] = M * (v1 - new_v1[i]) + v2;
			// ideally, we'd test for inequality.  however, i want
			// to avoid floating-point roundoff problems.  Compare
			// difference.
			if (fabs(KE
			         - (Obj_1->mass()*new_v1[i]*new_v1[i]
				     + Obj_2->mass()*new_v2[i]*new_v2[i]))
			    > ZERO_THRESHOLD ) {
				// The original eq. didn't work, therefore
				// we need to use the second value returned.
				fprintf(stderr, "Using second value from quad. eq.\n");
				new_v1[i] = x[1];
				new_v2[i] = M * (v1 - new_v1[i]) + v2;
			} else {
				fprintf(stderr, "Using first value from quad. eq.\n");
			}

		} else {
			// the answer is complex.  for now just barf
			fprintf(stderr, "Complex result for the %d dimension!!\n", i);
			fprintf(stderr, "possible new velocities: %g, %g\n",
			        x[0], x[1]);
			return false;

		} // Solve_Quadratic
		fprintf(stderr, "Dim %d: (v1',v2') = (%g, %g)\n", i, new_v1[i],
		        new_v2[i]);
	} // loop through three dimensions

	// finally, assign the new velocities to the objects
	Obj_1->setVelocity(vector(new_v1[0], new_v1[1], new_v1[2]));
	Obj_2->setVelocity(vector(new_v2[0], new_v2[1], new_v2[2]));

	Obj_1->velocity().array();
	fprintf(stderr, "Set: v1' = (%g, %g, %g)\n", Result[0],
	        Result[1], Result[2]);
	Obj_2->velocity().array();
 	fprintf(stderr, "Set: v2' = (%g, %g, %g)\n", Result[0],
		Result[1], Result[2]);

//	exit(1);
	return true;

}
*/

// Solve the quadratic equation.  Put both answers into the result array
// given in the fourth argument.  Return true for both answers real-valued,
// false for complex.  In the case of complex solutions return only the
// real component.
static bool Solve_Quadratic(scalar a, scalar b, scalar c, scalar x[2])
{
	bool Return_Val = true;
	scalar Discriminant = (b*b - 4 * a * c);

	if (fabs(Discriminant) < ZERO_THRESHOLD) {
		// tiny discriminant, set to zero to hold down errors.
		Discriminant = 0.0;
	} else if (Discriminant < ZERO_THRESHOLD) {
		// negative discriminant, complex answers.  If we set the
		// disc. to zero, we get the real component.
		Discriminant = 0.0;
		Return_Val = false;
	}

	/* assert( Discriminant >= 0.0);*/
	fprintf(stderr, "Discriminant: %g\n", Discriminant);
	fprintf(stderr, "(a,b,c) = (%g,%g,%g)\n", a, b, c);

	x[0] = (-b + sqrt(Discriminant)) / (2*a);
	if (fabs(x[0]) < ZERO_THRESHOLD) {
		x[0] = 0.0;
	}

	x[1] = (-b - sqrt(Discriminant)) / (2*a);
	if (fabs(x[1]) < ZERO_THRESHOLD) {
		x[1] = 0.0;
	}

	return Return_Val;
}