/* gravity_util.cc - various and sundry functions for dealing with gravity.
        Gravity is automatically applied to each object upon creation.
   When an object is removed, it ceases to effect the others
   gravitationally.  During a universe's lifetime, gravity will not
   automatically recalculate itself.  You should recalculate gravity
   whenever dealing with huge masses and huge distances (i.e, orbital-type
   stuff).  In terrestrial physics, gravity is generally constant.
         Also, you can fiddle with the gravitational constant, normally
   6.67258e-11 N m^2 / kg^2. (That's reallly tiny.)  Large values of G will
   need more frequent recalculations.
*/

#include "physics.h"
#include <stddef.h>

scalar universe::G(void) const
{
	if ((_object_list == NULL) || (_object_list->_gravity == NULL)) {
		// No objects in the universe... will use default value
		return UNIVERSAL_GRAVITATIONAL_CONSTANT;
	}

	// Only one `G' for every gravitational force	
	gravitation* Temp = dynamic_cast <gravitation*>
	                   (_object_list->_gravity->data);
	
	return Temp->G();
}

bool universe::setG(scalar New_G)
{
	if ((_object_list == NULL) || (_object_list->_gravity == NULL)) {
		// No objects in the universe... will use default value
		return false;
	}
	
	// Setting a single `G' sets property for all links.
	gravitation* Temp = dynamic_cast <gravitation*>
	                   (_object_list->_gravity->data);
	
	Temp->setG(New_G);
	
	return true;
	
}

// Here we want to recalculate every gravitational link for every object.
// Fun!
bool universe::calculateGravity(void)
{
	if (_object_list == NULL)
		return false;
	
	_on_every_object(_object_list, Refresh_Gravity);
	return true;
}

void Refresh_Gravity(point_mass* Object)
{
	if (Object == NULL)
		return;
		
	interaction_node* Here = Object->_gravity;
	while (Here != NULL) {
		Here->data->calculate();
		Here = Here->next;
	}
	return;
}

