/* interacton.cc - descriptions for different interations.
        Many physical laws can be described as interactions between two
   objects.  Actually, I think every law can...  Anyway, you create a law by
   linking two objects together.  This will add a constant force to each
   object.  Whenever the force changes (i.e., an object moves or changes
   state somehow), it's time to calculate(), changing said forces.
*/

#include "physics.h"

// For NULL
#include <stddef.h>

// debug
#include <stdio.h>

// abstract interactions.

interaction::interaction(void)
{
	_obj_1 = _obj_2 = NULL;
	_force_1 = _force_2 = (force_id) NULL;
}

// Creating a zero force gives us a handle to a force object.
// This way we can insure only changing that force in the future.
interaction::interaction(object* One, object* Two)
{
	_obj_1 = One;
	_obj_2 = Two;
	
	_force_1 = _force_2 = (force_id) NULL;
	
	
	if (_obj_1 != NULL) {
		_force_1 = _obj_1->applyForce(ZERO_VECTOR);
	}
	if (_obj_2 != NULL) {
		_force_2 = _obj_2->applyForce(ZERO_VECTOR);
	}
	
}

bool interaction::link(object* One, object* Two)
{
 	if ((One == NULL) || (Two == NULL)) {
 		return false;
 	}
 	_obj_1 = One;
 	_obj_2 = Two;
 	_force_1 = _obj_1->applyForce(ZERO_VECTOR);
 	_force_2 = _obj_2->applyForce(ZERO_VECTOR);
 	return true;
 	
}

interaction::~interaction(void)
{
	if ((_force_1 != (force_id) NULL) && (_obj_1 != NULL)) {
		_obj_1->removeForce(_force_1);
	}
	
	if ((_force_2 != (force_id) NULL) && (_obj_2 != NULL)) {
		_obj_2->removeForce(_force_2);
	}
	
}


// Gravitational relationships
scalar gravitation::_G = UNIVERSAL_GRAVITATIONAL_CONSTANT;

gravitation::gravitation(object* One, object* Two)
   : interaction(One, Two)
{
	if ((One != NULL) && (Two != NULL)) {
		_force_1 = One->applyForce(ZERO_VECTOR);
		_force_2 = Two->applyForce(ZERO_VECTOR);
		calculate();
	}
}

void gravitation::calculate(void)
{
	if ((_obj_1 == NULL) || (_obj_2 == NULL)) {
		return;
	}
	
	vector r = _obj_2->position() - _obj_1->position();
	
	scalar F = (_G * (_obj_1->mass()) * (_obj_2->mass()))
	           / (r.mag() * r.mag());
	
	
	// If separation is the zero vector, we'll get a divide-by-zero
	// problem.
	if (r.mag() != 0.0) {
		_obj_1->setForce(_force_1, F * r.dir());
		_obj_2->setForce(_force_2, F * -(r.dir()));
	} else {
		_obj_1->setForce(_force_1, ZERO_VECTOR);
		_obj_2->setForce(_force_2, ZERO_VECTOR);
	}

	return;
}

scalar gravitation::G(void) const
{
	return _G;
}

bool gravitation::setG(scalar G)
{
	_G = G;
	return true;
}


// Hooke's law - elasticity.  An object deformed will return to its regular
// state... restoring force is proportional to the distance the objects
// are apart.

// Constructor for spring not inside a universe
hookes_law::hookes_law(object* One, object* Two,
                       scalar k, scalar x, scalar max_x)
   : interaction()
{
	_k = k;
	_eq_point = x;
	_limit = max_x;
	_parent_uni = NULL;
	_id = 0;
}	
	
hookes_law::hookes_law(class universe* Parent, spring_id Id,
                       scalar k, scalar x, scalar max_x)
   : interaction()
{
	_k = k;
	_eq_point = x;
	_limit = max_x;
	_parent_uni = Parent;
	_id = Id;
}

hookes_law::hookes_law(class universe* Parent, spring_id Id,
                       object* One, object* Two,
                       scalar k, scalar x, scalar max_x)
   : interaction(One, Two)
{
	_k = k;
	_eq_point = x;
	_limit = max_x;
	_parent_uni = Parent;
	_id = Id;
	
	if ((One != NULL) && (Two != NULL)) {
		_force_1 = One->applyForce(ZERO_VECTOR);
		_force_2 = Two->applyForce(ZERO_VECTOR);
		calculate();
	}
}

hookes_law::~hookes_law(void)
{
	if (_parent_uni != NULL) {
		_parent_uni->springUnLink(_id);
	}
}

void hookes_law::calculate(void)
{
	
	if ((_obj_1 == NULL) || (_obj_2 == NULL)) {
		return;
	}
	
	vector r = _obj_2->position() - _obj_1->position();
	
	if (_limit > 0.0) {
		if ((r.mag() - _eq_point) > _limit) {
			delete this;
			return;
		}
	}
	
	scalar F = _k * (r.mag() - _eq_point);

	// If the distance between the two objects is 0, the direction
	// is undefined.
	if (r.mag() != 0.0) {
		_obj_1->setForce(_force_1, F * r.dir());
		_obj_2->setForce(_force_2, F * -(r.dir()));
	} else {
		_obj_1->setForce(_force_1, ZERO_VECTOR);
		_obj_2->setForce(_force_2, ZERO_VECTOR);
	}
}

scalar hookes_law::k(void) const
{
	return _k;
}

bool hookes_law::set_k(scalar k)
{
	_k = k;
	return true;
}

scalar hookes_law::eq(void) const
{
	return _eq_point;
}

bool hookes_law::setEq(scalar x)
{
	_eq_point = x;
	return true;
}

bool hookes_law::setId(spring_id Id)
{
	_id = Id;
	return true;
}

scalar hookes_law::limit(void) const
{
	return _limit;
}

bool hookes_law::setLimit(scalar max_x)
{
	_limit = max_x;
	return true;
}

// Friction! ***********************************
// not sure how scientific these equations are, but they're extensible
// and useful.

damping::damping(scalar c, scalar b, scalar a)
   : interaction()
{
	_a = a;
	_b = b;
	_c = c;
	
}

damping::damping(object* One, scalar c, scalar b, scalar a)
   : interaction(One, NULL)
{
	_a = a;
	_b = b;
	_c = c;
}

// Return a resistance force, F = -(a*v^2 + b*v + c)
void damping::calculate(void)
{
	if (_obj_1 == NULL) {
		return;
	}
	
	vector V = _obj_1->velocity();
	scalar V_mag = V.mag();
	scalar F_mag = (_a*V_mag*V_mag + _b*V_mag + _c);
	if ((F_mag == 0.0 || V_mag == 0.0)) {
		_obj_1->setForce(_force_1, ZERO_VECTOR);
	} else {
		_obj_1->setForce(_force_1, -F_mag * V.dir());
	}
}

vector damping::coeff(void)
{
	return ToVector(_a, _b, _c);
}

bool damping::setCoeff(vector K)
{
	_a = K.x();
	_b = K.y();
	_c = K.z();
	
	return true;
}

// Collision detection.
void collision::calculate(void)
{
	// Are static variables shared between instantiations of classes?
	// Better to put these in the class def.
	// Can't include a "point_mass" in interaction.h... circular dependencies
	// with physics.h.  I think the whole interaction system is b0rken.
	// Keep it as it was.  Could solve by making Mass_1 & Mass_2 automatic...
	// sloooooow.
	static point_mass* Mass_1 = dynamic_cast <point_mass*> (_obj_1);
	static point_mass* Mass_2 = dynamic_cast <point_mass*> (_obj_2);
//	static scalar Last_Collision = _parent_uni->time();

	if ((Mass_1 == NULL) || (Mass_2 == NULL)
	    || (_parent_uni == NULL)) {
		return;
	} else if ((Mass_1->radius() == 0.0)
	           || (Mass_2->radius() == 0.0)) {
		return;
	}

	// We'll give each object a tenth of a second to get out of the
	// collision zone.  This will hopefully reduce crazy bouncing
/*	scalar Now = _parent_uni->time();*/

	if ((_parent_uni->time() - Last_Collision) < 0.25) {
		return;
	}

	Last_Collision = _parent_uni->time();

	vector S = (Mass_1->position() - Mass_2->position());
	if (S.mag() < (Mass_1->radius() + Mass_2->radius())) {
		_parent_uni->collide(Mass_1->id(), Mass_2->id());
	}
}

collision::collision(universe* This)
   : interaction()
{
	_parent_uni = This;
	if (This != NULL) {
		Last_Collision = _parent_uni->time();
  	}
}

collision::collision(universe* This, object* One, object* Two)
   : interaction(One, Two)
{
	_parent_uni = This;
	if (This != NULL) {
		Last_Collision = _parent_uni->time();
  	}
//	Mass_1 = dynamic_cast <point_mass*> (_obj_1);
//	Mass_2 = dynamic_cast <point_mass*> (_obj_2);

}