/* population.cc -- methods for filling/emptying the universe of objects.
        Objects need to be created and destroyed.  Each should be given a
   unique id number to help keep track of things.
*/

#include "physics.h"

// for NULL
#include <stddef.h>

point_mass* Find_Tail(point_mass* Here);
interaction_node* Find_Tail(interaction_node* Here);
void Delete_Object(point_mass* Here);

object_id universe::add(scalar M, vector S, vector V, scalar r)
{
	object_id Last_Id = 0;
	
	point_mass* Tail = Find_Tail(_object_list);
	if (Tail == NULL) {
		_object_list = new point_mass("", M, S, V, r);
		Tail = _object_list;
	} else {
		Tail->_next = new point_mass("", M, S, V, r);
		Last_Id = Tail->_id;
		if (Tail->_next != NULL) {
			Tail->_next->_prev = Tail;
		}
		Tail = Tail->_next;
	}
	
	// was creation successful?
	if (Tail == NULL) {
		return 0;
	}
	
	// Add a new ID #....
	// this technique limits the universe to 2^15 objects or something.
	// this should be enough.
	
	Tail->_id = Last_Id + 1;
	
	// Set time attributes
	Tail->setTimeMult(_time_x);
	if (_paused) {
		Tail->setPause(_paused);
	}
	
	// Add a force for friction (damping)
	Tail->_friction = new damping(Tail, _damping_coeff[2],
	                              _damping_coeff[1], _damping_coeff[0]);
	Tail->_friction->calculate();
	
	// Finally, add a gravitation link to every object already in the
	// universe.
	// Also, add a collision detection interaction
	point_mass* Here = _object_list;
	while (Here != NULL) {
		if (Here != Tail) {
			Gravity_Link(Tail, Here);
			Collision_Link(this, Tail, Here);
		}
		Here = Here->_next;
	}
	
	return Tail->_id;
}

// Add a new link to the gravity list of the newest object.
void Gravity_Link(point_mass* New_Obj, point_mass* Existant)
{
	if ((New_Obj == NULL) || (Existant == NULL)) {
		return;
	}
	
	interaction_node* Tail = Find_Tail(New_Obj->_gravity);
	
	if (Tail == NULL) {
		New_Obj->_gravity = new interaction_node;
		Tail = New_Obj->_gravity;
	} else {
		Tail->next = new interaction_node;
		Tail = Tail->next;
	}
	
	if (Tail == NULL) {
		return;
	}
	
	Tail->data = new gravitation(New_Obj, Existant);
	Tail->members[0] = New_Obj->_id;
	Tail->members[1] = Existant->_id;
	Tail->next = NULL;
	
	return;
}

void Collision_Link(universe* Uni, point_mass* New_Obj,
                    point_mass* Existant)
{
	if ((New_Obj == NULL) || (Existant == NULL)) {
		return;
	}
	
	interaction_node* Tail = Find_Tail(New_Obj->_object_borders);
	
	if (Tail == NULL) {
		New_Obj->_object_borders = new interaction_node;
		Tail = New_Obj->_object_borders;
	} else {
		Tail->next = new interaction_node;
		Tail = Tail->next;
	}
	
	if (Tail == NULL) {
		return;
	}
	
	Tail->data = new collision(Uni, New_Obj, Existant);
	Tail->members[0] = New_Obj->_id;
	Tail->members[1] = Existant->_id;
	Tail->next = NULL;
	
	return;
}

point_mass* Find_Tail(point_mass* Here)
{
	if ((Here == NULL) || (Here->next() == NULL)) {
		return Here;
	} else {
		return Find_Tail(Here->next());
	}
}

interaction_node* Find_Tail(interaction_node* Here)
{
	if ((Here == NULL) || (Here->next == NULL)) {
		return Here;
	} else {
		return Find_Tail(Here->next);
	}
}

bool universe::remove(object_id Target)
{
	point_mass* Obj = _get_by_id(Target);
	if (Obj == NULL) {
		return false;
	}
	
	if (Obj == _object_list) {
		_object_list = Obj->_next;
	}
	
	delete Obj;
	return true;
}

bool universe::removeAll(void)
{
	_on_every_object(_object_list, Delete_Object);
	_object_list = NULL;
	
	return true;
}

// Simple way to tell if an ID # is valid.
bool universe::exists(object_id Id) const
{
	if (_get_by_id(Id) != NULL) {
		return true;
	} else {
		return false;
	}
}

char* universe::tag(object_id Target)
{
	point_mass* Obj = _get_by_id(Target);
	if (Obj == NULL) {
		return NULL;
	} else {
		return Obj->name();
	}
}

bool universe::setTag(object_id Target, char* Tag)
{
	point_mass* Obj = _get_by_id(Target);
	if (Obj == NULL) {
		return false;
	} else {
		return Obj->setName(Tag);
	}
}

void* universe::userData(object_id Target)
{
	point_mass* Obj = _get_by_id(Target);
	if (Obj == NULL) {
		return NULL;
	} else {
		return Obj->userData();
	}
}

bool universe::setUserData(object_id Target, void* Data)
{
	point_mass* Obj = _get_by_id(Target);
	if (Obj == NULL) {
		return false;
	} else {
		return Obj->setUserData(Data);
	}
}