/* springs.cc -- elastic connections between objects.
        Springs are good ways to make objects keep their relationships
   together.  An aggregate object could be created by linking together
   objects using springs with large spring constants.
        Since springs are not intrinsic to an object like gravity, the
   universe keeps track of them.  When an object is removed, any attached
   springs should also be removed.
*/
#include "physics.h"
#include <stddef.h>

interaction_node* Find_Tail(interaction_node* Here);
bool Delete_By_Id(interaction_node* Here, spring_id Id);
void Refresh_Spring(interaction_node* Spring);
interaction_node* Find_Spring(interaction_node* Here, spring_id Id);

int Count_Springs(interaction_node* Here);

spring_id universe::springLink(object_id Id_one, object_id Id_two,
                     scalar k, scalar x, scalar max_x)
{
	point_mass* One = _get_by_id(Id_one);
	point_mass* Two = _get_by_id(Id_two);
	
	if ((One == NULL) || (Two == NULL)) {
		return 0;
	}
	
	spring_id Last_Id = 0;
	
	interaction_node* Tail = Find_Tail(_spring_list);
	
	if (Tail == NULL) {
		_spring_list = new interaction_node;
		Tail = _spring_list;
	} else {
		Tail->next = new interaction_node;
		Last_Id = Tail->id;
		Tail = Tail->next;
	}
	
	if (Tail == NULL) {
		return 0;
	}
	
	Tail->id = Last_Id + 1;
	Tail->members[0] = One->_id;
	Tail->members[1] = Two->_id;
	Tail->next = NULL;
	Tail->data = new hookes_law(this, Tail->id, One, Two, k, x, max_x);
	
	return Tail->id;
	
}

bool universe::springUnLink(spring_id Target)
{
	// Check if there is only one spring.
	if ((_spring_list != NULL) && (_spring_list->next == NULL)) {
		delete _spring_list;
		_spring_list = NULL;
	}
	
	return Delete_By_Id(_spring_list, Target);
}

bool Delete_By_Id(interaction_node* Here, spring_id Id)
{
	if ((Here == NULL) || (Here->next == NULL)) {
		return false;
	} if (Here->next->id == Id) {
		interaction_node* Temp = Here->next;
		Here->next = Temp->next;
		delete Temp;
		return true;
	} else {
		return Delete_By_Id(Here->next, Id);
	}
}

// Calculate all springs in the universe.
bool universe::calculateSpring(void)
{
	if (_spring_list == NULL) {
		return false;
	}
	
	_on_every_object(_spring_list, Refresh_Spring);
	return true;
}

// calculate for single, named spring.
bool universe::calculateSpring(spring_id Target)
{
	if (_spring_list == NULL) {
		return false;
	}
	
	Refresh_Spring(Find_Spring(_spring_list, Target));
	return true;
}

void Refresh_Spring(interaction_node* Spring)
{
	if (Spring != NULL) {
		Spring->data->calculate();
	}
}

interaction_node* Find_Spring(interaction_node* Here, spring_id Id)
{
	while (Here != NULL) {
		if (Here->id == Id) {
			return Here;
		}
		Here = Here->next;
	}
	
	return NULL;
}

multi_spring* universe::springs(void)
{
	int Num_Springs = Count_Springs(_spring_list);
	multi_spring* Result = new multi_spring[Num_Springs + 1];
	interaction_node* Here = _spring_list;
	int i = 0;
	
	for (; i < Num_Springs; i++) {
		Result[i].data = dynamic_cast <hookes_law*> (Here->data);
		Result[i].one = Here->members[0];
		Result[i].two = Here->members[1];
		Here = Here->next;
	}
	
	// Terminate the end
	Result[i].data = NULL;
	Result[i].one = Result[i].two = 0;
	
	return Result;
	
}

int Count_Springs(interaction_node* Here)
{
	static int i = 0;
	
	if (Here == NULL) {
		int Temp = i;
		i = 0;
		return Temp;
	} else {
		i++;
		return Count_Springs(Here->next);
	}
}