/* timing.cc -- timers, clock multipliers, pause
        One of the more interesting aspects of playing with a physics engine
   would be fiddling with the rate of time.  Each object in the RockSpace
   universe keeps its own internal rate of time flow.  When you change the
   global time multiplier, all pre-existing objects use it and any freshly
   created objects will also use it.   Change the multiplier individually by
   supplying an id.
        Like the kinematics methods, the xxxPause() family of methods is
   overloaded to work on either indivindual objects or the universe at
   large.
*/

#include "physics.h"
#include <stddef.h>

#ifdef WIN32
// for ftime()
#include <sys/timeb.h>
#else     // not WIN32
// for getttimeofday()
#include <sys/time.h>
#include <unistd.h>
#endif    // WIN32

bool Change_Time_X(point_mass* Here, void* Arg);
bool Pause_Item(point_mass* Here, void* Arg);
void Toggle_Item(point_mass* Here);

// A global timekeeper would be useful, no?
scalar universe::time(void) const
{
	return (_time_x * (_now() - _initial_t));
}

scalar universe::_now(void) const
{
#ifndef WIN32
	struct timeval Now;
	gettimeofday(&Now, NULL);
	
	return (Now.tv_sec + (1e-6 * Now.tv_usec));
#else
	struct timeb Now;
	ftime(&Now);
	return (Now.time + (1e-3 * Now.millitm));
#endif    //  not WIN32		

}

// Return the global time multiplier.
scalar universe::timeMult(void) const
{
	return _time_x;
}

// If the object doesn't exist, return the global value.  (If an object
// with that id were created, it would inherit from global.)
scalar universe::timeMult(object_id Id) const
{
	point_mass* Target = _get_by_id(Id);
	if (Target == NULL) {
		return _time_x;
	} else {
		return Target->timeMult();
	}
}

bool universe::setTimeMult(scalar New_X)
{
	_time_x = New_X;
	// Now set the time multiplier for every object
	// We'll be kind of tricky... we'll send _search() a function that
	// never comes true, therefore working on every object in the list.
	// This way, we can send a parameter.
	_search(_object_list, (void*) &New_X, Change_Time_X);
	return true;
}

bool Change_Time_X(point_mass* Here, void* Arg)
{
	return Here->setTimeMult(*((scalar*) Arg));
}

bool universe::setTimeMult(object_id Id, scalar New_X)
{
	point_mass* Target = _get_by_id(Id);
	if (Target == NULL) {
		return false;
	} else {
		return Target->setTimeMult(New_X);
	}
}

// Now functions for (un)pausing the whole shebang. (and individual members)
bool universe::pause(void) const
{
	return _paused;
}

bool universe::pause(object_id Id) const
{
	point_mass* Target = _get_by_id(Id);
	if (Target == NULL) {
		return _paused;
	} else {
		return Target->pause();
	}
}

bool universe::setPause(bool Pause)
{
	_search(_object_list, (void*) &Pause, Pause_Item);
	
	return (_paused = Pause);
}

bool universe::setPause(object_id Id, bool Pause)
{
	point_mass* Target = _get_by_id(Id);
	if (Target != NULL) {
		return Pause_Item(Target, (void*) &Pause);
        }

	return false;
}

bool universe::togglePause(void)
{
	
	_on_every_object(_object_list, Toggle_Item);
	
	return (_paused = !_paused);
}

bool universe::togglePause(object_id Id)
{
	point_mass* Target = _get_by_id(Id);
	if (Target != NULL) {
		return Target->togglePause();
	}
	
	return false;
}

bool Pause_Item(point_mass* Here, void* Arg)
{
	return Here->setPause(*((bool*) Arg));
}

void Toggle_Item(point_mass* Here)
{
	Here->togglePause();
	return;
}