/* input.cc - methods and functions for converting a given RockSpace
      universe into XML
*/

#include "rs-xml.h"

#include <libxml/tree.h>

#include <stddef.h>
#include <string.h>
/* for snprintf */
#include <stdio.h>

#define STRING_BUFFER_LENGTH ( 256 )


// Local functions
static xmlDoc* Convert_Universe(universe* Uni);
static void Add_Objects(universe* Uni, xmlNode* Container);
static vector Lookup_By_Id(object_id Id, multi_vector* List);
static void Add_Vector_Child(xmlNode* Parent, char* Name, vector Data);
static void Add_Damping_Child(xmlNode* Parent, vector Data);
static void Add_Springs(universe* Uni, xmlNode* Container);
static void Add_Constants(universe* Uni, xmlNode* Container);

bool universe_xml_interface::input(universe* Uni)
{
	if (Uni == NULL) {
		return false;
	}
	
	// OK, clear out anything we had stored previously
	_clear_data();
	
	bool Previous_State = Uni->pause();
	
	Uni->setPause(true);
	
	_doc = Convert_Universe(Uni);
	
	if (_doc == NULL) {
		_clear_data();
		return false;
	}
	
	_root = xmlDocGetRootElement(_doc);
	
	if (_root == NULL) {
		_clear_data();
		return false;
	}
	
	Uni->setPause(Previous_State);
	
	return true;
}

static xmlDoc* Convert_Universe(universe* Uni)
{
	xmlDoc* Doc = xmlNewDoc((xmlChar*) "1.0");
	if (Doc == NULL) {
		return NULL;
	}
	
	Doc->children = xmlNewDocNode(Doc, NULL, (xmlChar*) "universe", NULL);

	Add_Objects(Uni, xmlNewChild(Doc->children, NULL, (xmlChar*) "objects",
	            NULL));
	Add_Springs(Uni, xmlNewChild(Doc->children, NULL, (xmlChar*) "springs",
	            NULL));
	Add_Constants(Uni, xmlNewChild(Doc->children, NULL,
	              (xmlChar*) "constants", NULL));
	return Doc;

}

static void Add_Objects(universe* Uni, xmlNode* Container)
{
	multi_vector* S = Uni->position();
	multi_vector* V = Uni->velocity();

	// S & V will always contain at least one item... the null terminator
	for (int i = 0; S[i].id != 0; i++) {
		xmlChar Buffer[STRING_BUFFER_LENGTH] = {0};
		
		xmlNode* Obj = xmlNewChild(Container, NULL, (xmlChar*) "object", NULL);
		
		// Set scalar properties
		snprintf((char*) Buffer, STRING_BUFFER_LENGTH, "%d", S[i].id);
		xmlNewProp(Obj, (xmlChar*) "id", Buffer);
		snprintf((char*) Buffer, STRING_BUFFER_LENGTH, "%g",
		         Uni->mass(S[i].id));
		
		// Add proper id #s to the xml/RockSpace table
		id_list* Newest = Add_Id_Entry();
		Newest->XML_id = Newest->RS_id = S[i].id;
		Newest->next = NULL;
		
		xmlNewProp(Obj, (xmlChar*) "mass", Buffer);
		snprintf((char*) Buffer, STRING_BUFFER_LENGTH, "%g",
		         Uni->radius(S[i].id));
		xmlNewProp(Obj, (xmlChar*) "radius", Buffer);
		xmlNewProp(Obj, (xmlChar*) "tag", (xmlChar*) Uni->tag(S[i].id));
		
		// Now, the vector properties are handled as children
		Add_Vector_Child(Obj, "position", S[i].data);
		Add_Vector_Child(Obj, "velocity", Lookup_By_Id(S[i].id, V));
		if (Uni->dampingCoeff() != Uni->dampingCoeff(S[i].id)) {
			Add_Damping_Child(Obj, Uni->dampingCoeff(S[i].id));
		}
	}
	
	delete [] S;
	delete [] V;

}

static vector Lookup_By_Id(object_id Id, multi_vector* List)
{
	for (int i = 0; List[i].id != 0; i++) {
		if (List[i].id == Id) {
			return List[i].data;
		}
	}
	
	return ZERO_VECTOR;
}

static void Add_Vector_Child(xmlNode* Parent, char* Name, vector Data)
{
	xmlNode* Vec = xmlNewChild(Parent, NULL, (xmlChar*) Name, NULL);
	if (Vec != NULL) {
		xmlChar Buffer[STRING_BUFFER_LENGTH];
		
		snprintf((char*) Buffer, STRING_BUFFER_LENGTH, "%g", Data.x());
		xmlNewProp(Vec, (xmlChar*) "x", Buffer);
		
		snprintf((char*) Buffer, STRING_BUFFER_LENGTH, "%g", Data.y());
		xmlNewProp(Vec, (xmlChar*) "y", Buffer);
		
		snprintf((char*) Buffer, STRING_BUFFER_LENGTH, "%g", Data.z());
		xmlNewProp(Vec, (xmlChar*) "z", Buffer);
	}
	
	return;
}

static void Add_Damping_Child(xmlNode* Parent, vector Data)
{
	xmlNode* Damp = xmlNewChild(Parent, NULL, (xmlChar*) "damping", NULL);
	if (Damp != NULL) {
		xmlChar Buffer[STRING_BUFFER_LENGTH];
		
		snprintf((char*) Buffer, STRING_BUFFER_LENGTH, "%g", Data.x());
		xmlNewProp(Damp, (xmlChar*) "a", Buffer);
		
		snprintf((char*) Buffer, STRING_BUFFER_LENGTH, "%g", Data.y());
		xmlNewProp(Damp, (xmlChar*) "b", Buffer);
		
		snprintf((char*) Buffer, STRING_BUFFER_LENGTH, "%g", Data.z());
		xmlNewProp(Damp, (xmlChar*) "c", Buffer);
	}
	return;
}

static void Add_Springs(universe* Uni, xmlNode* Container)
{
	multi_spring* Springs = Uni->springs();
	
	for (int i = 0; Springs[i].data != NULL; i++) {
		xmlChar Buffer[STRING_BUFFER_LENGTH] = {0};
		
		xmlNode* Obj = xmlNewChild(Container, NULL,
		               (xmlChar*) "spring", NULL);
		
		snprintf((char*) Buffer, STRING_BUFFER_LENGTH,
		         "%d", Springs[i].one);
		xmlNewProp(Obj, (xmlChar*) "id1", Buffer);
		snprintf((char*) Buffer, STRING_BUFFER_LENGTH,
		         "%d", Springs[i].two);
		xmlNewProp(Obj, (xmlChar*) "id2", Buffer);
		snprintf((char*) Buffer, STRING_BUFFER_LENGTH,
		         "%g", Springs[i].data->k());
		xmlNewProp(Obj, (xmlChar*) "k", Buffer);
		snprintf((char*) Buffer, STRING_BUFFER_LENGTH,
		         "%g", Springs[i].data->eq());
		xmlNewProp(Obj, (xmlChar*) "x", Buffer);
		snprintf((char*) Buffer, STRING_BUFFER_LENGTH,
		         "%g", Springs[i].data->limit());
		xmlNewProp(Obj, (xmlChar*) "max_x", Buffer);
		
	}
	
	delete [] Springs;
	
	return;
}

static void Add_Constants(universe* Uni, xmlNode* Container)
{
	xmlChar Buffer[STRING_BUFFER_LENGTH] = {0};
	xmlNode* Child = NULL;
	
	snprintf((char*) Buffer, STRING_BUFFER_LENGTH, "%g", Uni->G());
	Child = xmlNewChild(Container, NULL, (xmlChar*) "G", NULL);
	xmlNewProp(Child, (xmlChar*) "val", Buffer);
	
	snprintf((char*) Buffer, STRING_BUFFER_LENGTH, "%g", Uni->timeMult());
	Child = xmlNewChild(Container, NULL,
	        (xmlChar*) "time_multiplier", NULL);
	xmlNewProp(Child, (xmlChar*) "val", Buffer);
	
	vector D = Uni->dampingCoeff();
	Child = xmlNewChild(Container, NULL,
	        (xmlChar*) "damping", NULL);
	snprintf((char*) Buffer, STRING_BUFFER_LENGTH,
	         "%g", D.x());
	xmlNewProp(Child, (xmlChar*) "a", Buffer);
	
	snprintf((char*) Buffer, STRING_BUFFER_LENGTH,
	         "%g", D.y());
	xmlNewProp(Child, (xmlChar*) "b", Buffer);
	
	snprintf((char*) Buffer, STRING_BUFFER_LENGTH,
	         "%g", D.z());
	xmlNewProp(Child, (xmlChar*) "c", Buffer);
	
	return;

}

