/* load.cc - Load an XML file and convert it to a RockSpace universe.
*/

// 2/25/2001 - started work on updating for libxml 2.0

#include "rs-xml.h"

#include <libxml/parser.h>
#include <libxml/tree.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

// from universe_xml_interface.cc
void Delete_Nodes(id_list* Here);

// Local functions
static void XML_To_Universe(xmlNode* Root, universe* Uni);
static xmlNode* Find_Node_By_Name(xmlNode* Here, const char* Name);
static void Assign_Constants(xmlNode* Here, universe* Uni);
static void Build_Objects(xmlNode* Here, universe* Uni);
static void Connect_Springs(xmlNode* Here, universe* Uni);

// Allow member access to friend functions
static universe_xml_interface* Current = NULL;

bool universe_xml_interface::load(char* Filename)
{
	Current = this;
	
	// Clobber any pre-existing XML, universe
	_clear_data();
	
	// Now load the XML into memory
	_doc = xmlParseFile(Filename);
	if (_doc == NULL) {
		// bad file
		return false;
	}
	_root = xmlDocGetRootElement(_doc);
	if (_root == NULL) {
		// No information in XML
		_clear_data();
		return false;
	}
	
	// Create new universe to hold creation
	_uni = new universe;
	if (_uni == NULL) {
		// error allocating space
		_clear_data();
		return false;
	}
	_uni->setPause(true);
	
	// Fill universe
	XML_To_Universe(_root, _uni);
	
	return true;
	
}

	
void universe_xml_interface::_clear_data(void)
{
	if (_doc != NULL) {
		xmlFreeDoc(_doc);
	}
	
	if (_uni != NULL) {
		delete _uni;  // Default destructor
	}
	
	if (_id_table != NULL) {
		Delete_Nodes(_id_table);
	}
}

static void XML_To_Universe(xmlNode* Root, universe* Uni)
{
	// First set all the universal constants
        xmlNode* Constants = Find_Node_By_Name(Root, "constants");

        if (Constants != NULL) {
        	Assign_Constants(Constants->children, Uni);
        }

        // Now build objects
        xmlNode* Objects = Find_Node_By_Name(Root, "objects");

        if (Objects != NULL) {
        	Build_Objects(Objects->children, Uni);
        }

        // Finally, connect the objects with springs
        xmlNode* Springs = Find_Node_By_Name(Root, "springs");

        if (Springs != NULL) {
        	Connect_Springs(Springs->children, Uni);
        }

	return;
}

static xmlNode* Find_Node_By_Name(xmlNode* Here, const char* Name)
{
	if (Here == NULL) {
		return NULL;
	} else if (strcmp((char*) Here->name, Name) == 0) {
		return Here;
	} else {
		// Need to check two branches
		xmlNode* Temp = Find_Node_By_Name(Here->next, Name);
		if (Temp != NULL) {
			return Temp;
		} else {
			return Find_Node_By_Name(Here->children, Name);
		}
	}
}

static void Assign_Constants(xmlNode* Here, universe* Uni)
{
	if ((Here == NULL) || (Uni == NULL)) {
		return;
	}
	
	// Recursively traverse the rest of the list
	Assign_Constants(Here->next, Uni);
	
	// Ok, now examine this node
	if (strcmp((char*) Here->name, "damping") == 0) {
		Uni->setDampingCoeff(Extract_Vector(Here));
	} else if (strcmp((char*) Here->name, "G") == 0) {
		Uni->setG(atof(Extract_Attr(Here, "val")));
		
	} else if (strcmp((char*) Here->name, "time_multiplier") == 0) {
		Uni->setTimeMult(atof(Extract_Attr(Here, "val")));
	}
}

static void Build_Objects(xmlNode* Here, universe* Uni)
{
	// Enable recursion
	if ((Here == NULL) || (Uni == NULL)) {
		return;
	}
	Build_Objects(Here->next, Uni);
	
	// Down to business.  Fields to be filled
	int Xml_id = atoi(Extract_Attr(Here, "id"));
	char* Tag = Extract_Attr(Here, "tag");
	char* Extra = NULL;
	scalar M = atof(Extract_Attr(Here, "mass"));
	scalar r = atof(Extract_Attr(Here, "radius"));
	vector S = ZERO_VECTOR;
	vector V = ZERO_VECTOR;
	vector D = ZERO_VECTOR;
	
	if (strlen(Extract_Attr(Here, "userdata")) != 0) {
		Extra = new char[strlen(Extract_Attr(Here, "userdata")) + 1]; 	
		strcpy(Extra, Extract_Attr(Here, "userdata"));
	}
		
	// Now we need to search through the children for vector quantities
	xmlNode* Child = Here->children;
	while (Child != NULL) {
		if (strcmp((char*) Child->name, "position") == 0) {
			S = Extract_Vector(Child);
		} else if (strcmp((char*) Child->name, "velocity") == 0) {
			V = Extract_Vector(Child);
		} else if (strcmp((char*) Child->name, "damping") == 0) {
			D = Extract_Vector(Child);
		}
		
		Child = Child->next;
	}
	
	// Add space to store this object's id in the id table.
	id_list* Newest_Entry = Add_Id_Entry();
	if (Newest_Entry == NULL) {
		return;
	}
	
	Newest_Entry->XML_id = Xml_id;
	Newest_Entry->RS_id = Uni->add(M, S, V, r);
	Uni->setTag(Newest_Entry->RS_id, Tag);
	Uni->setUserData(Newest_Entry->RS_id, (void*) Extra);
	// Why is this check here?  Isn't default damping zero?
	// That may change.
	if (D != ZERO_VECTOR) {
		Uni->setDampingCoeff(Newest_Entry->RS_id, D);
	}
	
	return;
	
}



id_list* Add_Id_Entry(void)
{
	id_list* Here = Current->_id_table;
	
	if (Here == NULL) {
		// empty list
		Current->_id_table = Here = new id_list;
		
	} else {
		while (Here->next != NULL) {
			Here = Here->next;
		}
		Here->next = new id_list;
		Here = Here->next;
	}
	
	Here->next = NULL;
	Here->RS_id = Here->XML_id = 0;
	return Here;
}

static void Connect_Springs(xmlNode* Here, universe* Uni)
{
	// Enable recursion
	if ((Here == NULL) || (Uni == NULL)) {
		return;
	}
	Connect_Springs(Here->next, Uni);
	
	// Down to business. Everything is an attribute.
	scalar k = DEFAULT_K,
	       x = DEFAULT_EQ_POINT,
	       max_x = DEFAULT_ELASTIC_LIMIT;
	object_id One = Current->rsId(atoi(Extract_Attr(Here, "id1"))),
	          Two = Current->rsId(atoi(Extract_Attr(Here, "id2")));

	// If the XML doesn't contain an attribute, we should
	// use the default.  Therefore, we check if attributes are
	// contained.
	scalar Temp = 0.0;
	Temp = atof(Extract_Attr(Here, "k"));
	if (Temp != 0.0) {
		k = Temp;
	}
	Temp = atof(Extract_Attr(Here, "x"));
	if (Temp != 0.0) {
		x = Temp;
	}
	Temp = atof(Extract_Attr(Here, "max_x"));
	if (Temp != 0.0) {
		max_x = Temp;
	}
	
	if ((One == 0) || (Two == 0)) {
		return;
	}
	
	Uni->springLink(One, Two, k, x, max_x);
	return;
	
}

